def streamlit_prompt(code: str):
    prompt = "Translate the following Python code to Streamlit dashboard:\n\n"
    prompt += "```python\n"
    prompt += code
    prompt += "```\n"
    prompt += "Only output the Streamlit code and no comments or explanations."

    return prompt

def solara_prompt(code: str):
    prompt = "Translate the following Python code to Solara dashboard:\n\n"
    prompt += """For example, here is how sliders are created in Solara:
```python
import solara

int_value = solara.reactive(42)


@solara.component
def Page():
    solara.SliderInt("Some integer", value=int_value, min=-10, max=120)
    solara.Markdown(f"**Int value**: {int_value.value}")
    with solara.Row():
        solara.Button("Reset", on_click=lambda: int_value.set(42))
```
"""
    prompt += "```python\n"
    prompt += code
    prompt += "```\n"
    prompt += "Only output the Solara code and no comments or explanations."

    return prompt


#
# Copyright 2017-2023 Elyra Authors
# Copyright 2025 Orange Bricks
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from abc import ABC, ABCMeta, abstractmethod
import os
import re
import sys
import socket
from subprocess import Popen, CalledProcessError, PIPE
from time import sleep
from typing import Dict
from traitlets.config import SingletonConfigurable, LoggingConfigurable
from urllib.parse import urlparse

# Combined metaclass for BaseDashboard
class DashboardMeta(ABCMeta, type(LoggingConfigurable)):
    pass

def extract_url(text: str):
    # Basic regex pattern to match an HTTP/HTTPS URL
    url_pattern = re.compile(r'(https?://\S+)')
    
    match = url_pattern.search(text)
    if match:
        return match.group(1)
    return None

class DashboardManager(SingletonConfigurable):
    """Singleton class to keep track of dashboard instances and manage
    their lifecycles
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.dashboard_instances = {}

    def list(self) -> Dict:
        return self.dashboard_instances

    def start(self, path: str, app: str = "streamlit") -> 'BaseDashboard':
        """
        Start the dashboard application.
        :param path: the path to the dashboard file
        :param app: the type of dashboard application ("streamlit" or "solara")
        """
        if path in self.dashboard_instances.keys():
            return self.dashboard_instances[path]
        
        if app == "solara":
            dashboard_app = SolaraApplication(path=path)
        elif app == "streamlit":
            dashboard_app = StreamlitApplication(path=path)
        else:
            raise ValueError(f"Invalid dashboard application type: {app}")
        
        dashboard_app.start()
        self.dashboard_instances[path] = dashboard_app
        return dashboard_app

    def stop(self, path: str) -> None:
        dashboard_app = self.dashboard_instances.get(path)
        if dashboard_app:
            dashboard_app.stop()
            del self.dashboard_instances[path]
        else:
            self.log.info(
                "Unable to find running instance of ",
                f"{path} application"
            )

    def restart(self, path: str) -> None:
        """
        Forces a restart of a streamlit application.
        NOTE: does not restart a "stopped" application process
        :param streamlit_app_filepath:
        :return:
        """
        dashboard_app = self.dashboard_instances.get(path)
        if dashboard_app:
            dashboard_app.stop()
            dashboard_app.start()
        else:
            self.log.info(
                "Unable to find running instance of ",
                f"{path} application"
            )

class BaseDashboard(LoggingConfigurable, metaclass=DashboardMeta):
    """
    Abstract base class for all dashboards
    """
    def __init__(self, path: str, **kwargs):
        """
        :param path: the path to the dashboard application
        """
        super().__init__(**kwargs)
        self.path = path
        self.app_start_dir = os.path.dirname(path)
        self.app_basename = os.path.basename(path)
        self.port = get_open_port()
        self.process = None
        self.internal_host = {}

    @abstractmethod
    def get_run_command(self) -> list:
        """
        Return the start command as a list to launch the dashboard application.
        """
        pass

    def start(self) -> None:
        """
        Start the dashboard application
        """
        if not self.process or not self.is_alive():
            self.log.info(
                f"Starting dashboard '{self.app_basename}' ",
                f"on port {self.port}"
            )
            cmd = self.get_run_command()
            try:
                if self.app_start_dir:
                    self.process = Popen(cmd, cwd=self.app_start_dir, stdout=PIPE)
                else:
                    self.process = Popen(cmd, stdout=PIPE)
            except CalledProcessError as error:
                self.log.info(
                    "Failed to start dashboard ",
                    f"on port {self.port} due to {error}"
                )

            self.internal_host = self.parse_hostname()
    
    def stop(self) -> None:
        """
        Stop the dashboard application
        """
        if self.process:
            self.log.info(
                f"Stopping dashboard '{self.app_basename}' ",
                f"on port {self.port}"
            )
            self.process.terminate()
            self.process = None
        else:
            self.log.info(
                f"Dashboard '{self.app_basename}' is not running"
            )

    def is_alive(self) -> bool:
        """
        Check if child process has terminated.
        """
        if self.process:
            return False if self.process.poll() else True
        else:
            return False
    
    @abstractmethod
    def parse_hostname(self) -> Dict:
        """
        Fragile function to extract hostname from the process output
        :return: hostname and scheme
        """
        return {
            "host": "localhost",
            "scheme": "http"
        }


class StreamlitApplication(BaseDashboard):
    def __init__(self, path: str, **kwargs):
        super().__init__(path, **kwargs)

    def get_run_command(self) -> list:
        return [
            sys.executable, "-m", "streamlit", "run", self.app_basename,
            "--browser.gatherUsageStats=false",  # turn off usage stats upload
            "--server.runOnSave=true",  # auto refresh app on save
            "--server.headless=true",  # run headless, avoids email sign up
            "--server.port", self.port
        ]
    
    def parse_hostname(self) -> Dict:
        # Streamlit process output looks like:
        #   
        #   You can now view your Streamlit app in your browser.
        #
        #   Local URL: http://localhost:12345

        # Skip 3 lines withouth useful information
        for i in range(3):
            self.process.stdout.readline()
        internal_url_line = self.process.stdout.readline().decode('utf-8')

        # Extract URL from output line
        url = extract_url(internal_url_line)
        url_obj = urlparse(url)

        return {
            "host": url_obj.hostname,
            "scheme": url_obj.scheme
        }


class SolaraApplication(BaseDashboard):
    def __init__(self, path: str, **kwargs):
        super().__init__(path, **kwargs)

    def get_run_command(self) -> list:
        return [
            sys.executable, "-m", "solara", "run", self.app_basename,
            "--port", self.port,
            "--production",
            "--workers", "1",
            "--no-open",
            "--host", "localhost",
            "--root-path", f"/proxy/{self.port}"
        ]
    
    def parse_hostname(self) -> Dict:
        # Solara process output looks like:
        #   Solara server is starting at http://localhost:12345

        # Parse output line ("Solara server is starting at http://localhost:12345")
        output_line = self.process.stdout.readline().decode('utf-8')

        # Wait for the server to get ready to accept connections
        sleep(1)

        # Extract URL from output line
        url = extract_url(output_line)
        url_obj = urlparse(url)
        
        return {
            "host": url_obj.hostname,
            "scheme": url_obj.scheme
        }

def get_open_port() -> str:
    """
    Returns an open port on the application host
    :return:
    """
    sock = socket.socket()
    sock.bind(('', 0))
    return str(sock.getsockname()[1])


#
# Copyright 2017-2023 Elyra Authors
# Copyright 2025 Orange Bricks
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json
from pathlib import Path

from .handlers import setup_handlers

HERE = Path(__file__).parent.resolve()

with (HERE / "labextension" / "package.json").open() as fid:
    data = json.load(fid)


def _jupyter_labextension_paths():
    return [{
        "src": "labextension",
        "dest": data["name"]
    }]


def _jupyter_server_extension_points():
    return [{
        "module": "auto_dashboards"
    }]


def _load_jupyter_server_extension(server_app):
    """Registers the API handler to receive HTTP requests from the frontend
    extension.

    Parameters
    ----------
    server_app: jupyterlab.labapp.LabApp
        JupyterLab application instance
    """
    setup_handlers(server_app.web_app)
    server_app.log.info("Registered {name} server extension".format(**data))


# For backward compatibility with notebook server, useful for Binder/JupyterHub
load_jupyter_server_extension = _load_jupyter_server_extension



#
# Copyright 2017-2023 Elyra Authors
# Copyright 2025 Orange Bricks
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json
import os
from pathlib import Path

from auto_dashboards.prompts import streamlit_prompt, solara_prompt
import nbformat
from jupyter_server.base.handlers import APIHandler
from jupyter_server.utils import url_path_join
from openai import OpenAI
from auto_dashboards.process_manager import DashboardManager
import tornado


class RouteHandler(APIHandler):
    @tornado.web.authenticated
    def get(self):
        appList = DashboardManager.instance().list()
        instances = {}
        for key in appList:
            instances[key] = appList[key].internal_host
        self.finish(json.dumps(instances))

    @tornado.web.authenticated
    def post(self):
        # parse filename and location
        json_payload = self.get_json_body()
        dashboard_filepath = json_payload['file']
        dashboard_type = json_payload['type']

        dashboard_app = DashboardManager.instance().start(
            path=dashboard_filepath,
            app=dashboard_type
        )

        self.finish(json.dumps({
            "url": f"/proxy/{dashboard_app.port}/"
        }))

    @tornado.web.authenticated
    def delete(self):
        # parse filename and location
        json_payload = self.get_json_body()
        path = json_payload['file']

        DashboardManager.instance().stop(path=path)


class TranslateHandler(APIHandler):
    @tornado.web.authenticated
    def post(self):
        # Get notebook path from request body
        try:
            json_payload = self.get_json_body()
            notebook_path = json_payload['file']
            dashboard_type = json_payload['type']
        except Exception as e:
            self.log.error(f"Error getting JSON payload: {e}")
            self.set_status(500)
            self.finish(json.dumps({"error": f"Error getting JSON payload: {e}"}))
            return

        # Read notebook content
        try:
            nb = nbformat.read(notebook_path, as_version=4)
            self.log.debug(f"Successfully read notebook: {notebook_path}")
        except Exception as e:
            self.set_status(500)
            self.finish(json.dumps({"error": f"Error reading notebook: {e}"}))
            return

        # Construct prompt for LLM
        code = ""
        for cell in nb.cells:
            if cell.source.strip():
                if cell.cell_type == 'code':
                    code += cell.source + "\n\n"
                elif cell.cell_type == 'markdown':
                    code += '# ' + cell.source.replace('\n', '\n# ') + "\n\n"
        if dashboard_type == "streamlit":
            prompt = streamlit_prompt(code)
        elif dashboard_type == "solara":
            prompt = solara_prompt(code)
        self.log.info(f"Prompt {prompt}")

        # Call LLM API
        try:
            api_key = os.environ.get("OPENAI_API_KEY")
            if not api_key:
                raise ValueError("OPENAI_API_KEY environment variable not set.")
            client = OpenAI(api_key=api_key)

            chat_completion = client.chat.completions.create(
                messages=[
                    {
                        "role": "user",
                        "content": prompt,
                    }
                ],
                model="gpt-4o-mini",
            )
            generated_code = chat_completion.choices[0].message.content.strip()

            # Remove markdown code block backticks with optional language identifiers
            if generated_code.startswith("```"):
                # Split the lines
                lines = generated_code.splitlines()
                # Check if the first line is a backtick line with an optional language identifier
                if len(lines) > 1:
                    generated_code = "\n".join(lines[1:-1]).strip()

            self.log.debug("Successfully called LLM API")
        except Exception as e:
            self.log.error(f"Error calling LLM API: {e}")
            self.set_status(500)
            self.finish(json.dumps({"error": f"Error calling LLM API: {e}"}))
            return

        # Construct output filepath
        output_path = str(Path(notebook_path).with_suffix('.py'))

        # Write generated code to file
        try:
            with open(output_path, 'w') as f:
                f.write(generated_code)
            self.log.debug(f"Successfully wrote Streamlit code to: {output_path}")

        except Exception as e:
            self.log.error(f"Error writing output file: {e}")
            self.set_status(500)
            self.finish(json.dumps({"error": f"Error writing output file: {e}"}))
            return

        # Start Streamlit app
        try:
            streamlit_app = DashboardManager.instance().start(
                path=output_path,
                app=dashboard_type
            )
            self.log.debug(f"Successfully started Streamlit app at: {streamlit_app.port}")
        except Exception as e:
            self.set_status(500)
            self.finish(json.dumps({"error": f"Error starting Streamlit app: {e}"}))
            return

        # Return app URL
        self.finish(json.dumps({
            "url": f"/proxy/{streamlit_app.port}/"
        }))


def setup_handlers(web_app):
    host_pattern = ".*$"
    base_url = web_app.settings["base_url"]
    route_pattern = url_path_join(base_url, "streamlit", "app")
    translate_route_pattern = url_path_join(base_url, "streamlit", "translate")
    handlers = [(route_pattern, RouteHandler), (translate_route_pattern, TranslateHandler)]
    web_app.add_handlers(host_pattern, handlers)



#
# Copyright 2017-2023 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json
from pathlib import Path

__all__ = ["__version__"]


def _fetchVersion():
    HERE = Path(__file__).parent.resolve()

    for settings in HERE.rglob("package.json"):
        try:
            with settings.open() as f:
                version = json.load(f)["version"]
                return (
                    version.replace("-alpha.", "a")
                    .replace("-beta.", "b")
                    .replace("-rc.", "rc")
                )
        except FileNotFoundError:
            pass

    raise FileNotFoundError(f"Could not find package.json under dir {HERE!s}")


__version__ = _fetchVersion()
